"""
All operations related to sentiment
"""
from relevanceai.operations_new.sentiment.base import SentimentBase
from relevanceai.operations_new.apibase import OperationAPIBase


class SentimentOps(SentimentBase, OperationAPIBase):
    pass
